"""Umap Splatters
To work you'll need:

pip install numba==0.50.1  # 0.51.2 might work, but 0.50.1 safer
pip install umap-learn
pip install matplotlib
pip install datashader
pip install bokeh
pip install pandas
pip install holoviews
pip install colorcet
"""

# from oui.util import ModuleNotFoundIgnore
# with ModuleNotFoundIgnore():
#     ...

try:
    import umap
    import umap.plot
    from numba.core.typing import cffi_utils

except Exception as err:
    print(f"If you get a TypeError: wrapped() missing 1 required positional argument: 'name', "
          f"you might want to consider doing a ``pip install numba==0.50.1``")
    raise


# from numba.core.typing import cffi_utils
# import pandas as pd

def plot_umap_of_xy(X, y=None, return_projection=False, **umap_plot_points_kwargs):
    model = umap.UMAP()
    model.fit(X)
    t = umap.plot.points(model, labels=y, **umap_plot_points_kwargs)
    if not return_projection:
        return t
    else:
        projection = model.transform(X)
        return t, projection



from oui import get_pkg_data
from typing import Iterable

alpha_less_rgb_hex_length = len('#aabbcc')


def int_to_alpha_hex(alpha: int = 255):
    assert isinstance(alpha, int) and 0 <= alpha <= 255, \
        f"int alpha should be between 0 and 255"
    return '%02x' % alpha


def float_to_alpha_hex(alpha: float = 1.0):
    assert 0 <= alpha <= 1, f"float alpha should be between 0 and 1"
    return int_to_alpha_hex(int_to_alpha_hex(int(alpha)))


def alpha_hex_from(alpha):
    if isinstance(alpha, int):
        return int_to_alpha_hex(alpha)
    elif isinstance(alpha, float):
        return float_to_alpha_hex(alpha)
    elif isinstance(alpha, str):
        return alpha


def color_hex_to_color_hex(color_hex):
    """
    >>> color_hex_to_color_hex('#3af')
    '#33aaff'
    >>> color_hex_to_color_hex('#09bd')
    '#0099bbdd'
    >>> color_hex_to_color_hex('#00ff1080')
    '#00ff1080'
    """
    n = len(color_hex)
    assert n > 0 and color_hex[0] == '#', "color hex code should start with #"
    assert n in {4, 5, 7, 9}, "color hex code should be # followed with 3, 4, 6, or 8 digits"
    if n <= 5:
        h = color_hex[0] + color_hex[1] * 2 + color_hex[2] * 2 + color_hex[3] * 2
        if n == 5:
            h += color_hex[4] * 2
        return h
    else:  # return as is
        return color_hex


def add_alpha(rgb_hex, alpha_hex, replace_existing_alpha=False):
    """Add an alpha to the rgb_hex.
    >>> add_alpha('#123456', 'ab')
    '#123456ab'
    >>> add_alpha('#12345678', 'ab')
    '#12345678'
    >>> add_alpha('#12345678', 'ab', replace_existing_alpha=True)
    '#123456ab'
    """
    if len(rgb_hex) > alpha_less_rgb_hex_length:
        if replace_existing_alpha:
            return rgb_hex[:-2] + alpha_hex
        else:
            return rgb_hex
    else:
        return rgb_hex + alpha_hex


color_names_and_codes = get_pkg_data('color_names_and_codes.json')
hex_for_color = {x['color']: x['hex'] for x in color_names_and_codes}


def dec_to_hex(dec):
    return '%02x' % dec


def color_dec_to_hex(color_dec):
    """Get the color hex code for an integer (r, g, b) or (r, g, b, alpha) specification.
    >>> color_dec_to_hex((10, 64, 200))
    '#0a40c8'
    >>> color_dec_to_hex((0, 255, 16, 128))
    '#00ff1080'
    """
    assert 3 <= len(color_dec) <= 4 and all(0 <= x <= 255 for x in color_dec), \
        f"Should be a triple with integers between 0 and 255. Was {color_dec}"
    return '#' + ''.join(map(dec_to_hex, color_dec))


def color_hex_from(color_spec):
    if isinstance(color_spec, str):
        if color_spec in hex_for_color:
            return hex_for_color[color_spec]
        else:
            return color_hex_to_color_hex(color_spec)  # just to normalize (expand #123 short-hands)
    elif isinstance(color_spec, Iterable) and 3 <= len(color_spec) <= 4:
        return color_dec_to_hex(color_spec)
    else:
        raise ValueError(f"color spec must be a recognized color, "
                         f"an hex color spec (starting with #) or a rgb triple. Was: {color_spec}")


class _HexColor:
    """A convenience class to access color hex codes from color names and short-hands"""

    def __init__(self):
        for _color, _color_hex in hex_for_color.items():
            setattr(self, _color.replace(' ', '_'), _color_hex)

    def __iter__(self):
        yield from hex_for_color


hex_color = _HexColor()



from functools import partial
from IPython.display import HTML, Javascript
import os
from i2.deco import postprocess
import json
from collections.abc import Mapping
from math import sqrt, pi
from pathlib import Path

from i2.signatures import Sig
from oui.color_util import color_hex_from, add_alpha, dec_to_hex

HTML('<script>var exports = {"__esModule": true};</script>')

pkg_dir = os.path.dirname(__file__)
dflt_filename = 'splatter_defaults.json'
pjoin = lambda *p: os.path.join(pkg_dir, *p)
dflts_filepath = pjoin(dflt_filename)

dflts = json.load(open(dflts_filepath, 'r'))
splatter_dflts = dict(dflts['options'], **dflts['tsneOptions'])

_splatter_raw_sig = Sig.from_objs('pts', splatter_dflts.items())


# splatter_kwargs_dflts = {k: dflts[k] for k in dflts if k not in }
# _splatter_sig = Sig.from_objs('pts', splatter_dflts.items())


def assert_jsonizable(d):
    _ = json.dumps(d)
    return True

# @Sig.from_objs('pts', dflts.items(), assert_same_sized_fvs=True)
# def mysplatter():
#     pass

def process_pts(pts):
    """Get a normalize form for pts.

    Feed `fvs` (an iterable of individual fvs) to get the [{tag: fv}, ...] form out of it
    >>> list(process_pts(pts=[[1, 2, 3], [4, 5, 6]]))
    [{'fv': [1, 2, 3]}, {'fv': [4, 5, 6]}]

    Feed `{tag: fvs,...}` form to get the [{tag: fv}, ...] form out of it
    >>> list(process_pts(pts={
    ...     'old': [[1, 2], [3, 4], [5, 6]],
    ...     'new': [[10, 20], [30, 40]]}))  # doctest: +NORMALIZE_WHITESPACE
    [{'tag': 'old', 'fv': [1, 2]},
     {'tag': 'old', 'fv': [3, 4]},
     {'tag': 'old', 'fv': [5, 6]},
     {'tag': 'new', 'fv': [10, 20]},
     {'tag': 'new', 'fv': [30, 40]}]

    :param pts: Some form of pts
    :return: A normalize "list-of-dicts" form of pts, with dicts being at least
    """
    if not isinstance(pts, Mapping):
        # pts is a list(-like)
        for pt in pts:
            if isinstance(pt, Mapping):
                pt = dict(**pt)  # cast to dict (or make copy)
                pt['fv'] = list(pt['fv'])  # make sure fv is a list
            else:
                pt = {'fv': list(pt)}  # if pt not a dict, assume it's the fv itself
            yield pt
    else:  # if pts is a Mapping, it's a {tag: fvs} mapping:
        for tag, fvs in pts.items():
            for fv in fvs:
                yield dict(tag=tag, fv=list(fv))


_max_node_size_ratio = 0.20


def ordered_uniks(iterable):
    """Unique values of iterable in the order they are encountered in arr
    >>> iterable = [4, 2, 6, 1, 2, 2, 7]
    >>> ordered_uniks(iterable)
    [4, 2, 6, 1, 7]
    """
    found = set()
    # Note: (found.add(x) is None) is a trick so that it the expression is always evaluated.
    return [x for x in iterable if x not in found and found.add(x) is None]


def process_viz_args(pts, nodeSize, figsize, fillColors, untaggedColor, alpha=1):
    n = len(pts)
    if isinstance(figsize, (int, float)):
        figsize = (figsize, figsize)
    height, width = figsize

    dflt_fill_colors = splatter_dflts['fillColors']
    if fillColors is None:
        fillColors = dflt_fill_colors
    elif isinstance(fillColors, Mapping):
        color_for_tag = fillColors
        unik_tags = ordered_uniks(filter(None, (x.get('tag', None) for x in pts)))
        when_not_found_choose_from_here = iter(dflt_fill_colors)
        fillColors = [color_for_tag.get(tag, False) or next(when_not_found_choose_from_here) for tag in unik_tags]

        # If '' and None was mentioned in color_for_tag, the user wants to specify untaggedColor
        if '' in color_for_tag:
            untaggedColor = color_for_tag['']
        elif None in color_for_tag:
            untaggedColor = color_for_tag[None]

    if nodeSize < _max_node_size_ratio:  # if smaller than max_node_size_ratio, it's not pixels,
        # but a desired node coverage ratio (approx ratio of the figure coverage by nodes)
        # It assumes the (unknown) nodeSize is a the "radius" of a circle so that
        # node_coverage = n * pi * nodeSize ** 2 / (height * width)
        # And solves for nodeSize to get desired coverage.
        node_coverage = nodeSize
        nodeSize = max(1, sqrt(node_coverage * (height * width) / (pi * n)))

    assert 0 <= alpha <= 1, f"alpha should be between 0 and 1, was {alpha}"
    fillColors = list(map(color_hex_from, fillColors))
    untaggedColor = color_hex_from(untaggedColor)
    if alpha != 1:
        alpha_hex = dec_to_hex(round(int(alpha * 255)))
        add_this_alpha = partial(add_alpha, alpha_hex=alpha_hex)
        fillColors = list(map(add_this_alpha, fillColors))
        untaggedColor = add_this_alpha(untaggedColor)
    return pts, nodeSize, figsize, fillColors, untaggedColor


# TODO: Wishlist: A decorator to automatically make extra_splatter_kwargs explicit (from dflts)
def splatter(pts,
             nodeSize=0.02,
             figsize=(200, 200),
             fillColors=None,
             untaggedColor='#444',
             alpha=1,
             process_pts=process_pts,
             process_viz_args=process_viz_args,
             **extra_splatter_kwargs):
    pts = list(process_pts(pts))
    pts, nodeSize, figsize, fillColors, untaggedColor = process_viz_args(
        pts, nodeSize, figsize, fillColors, untaggedColor, alpha)
    height, width = figsize
    return splatter_raw(pts, nodeSize=nodeSize, height=height, width=width,
                        fillColors=fillColors, untaggedColor=untaggedColor,
                        **extra_splatter_kwargs)


# TODO: Forward JS errors to python and handle on python side (raising informative error for e.g.)
@_splatter_raw_sig
def splatter_raw(*args, **kwargs):
    """
    Splatter multidimensional pts (that is, see a TSNE iteration happen in front of your eyes,
    squishing your multidimensional pts into two dimensions.

    The `pts` input is a list of dicts, where every dict must have, at a minimum, an `'fv'` field whose value
    is a list of fixed size for all pt of pts.

    Optionally, you can include:
    - 'tag': Will be use to categorize and color the point

    :param pts: Your pts, in the form of a list of dicts, list of lists, or dict of lists.
        All forms of data will be converted to a list of dicts where these dicts have four fields
        (other fields are possible, but are ignored by splatter): `fv` (required), `tag`, `source`, and `bt`.
        - `fv`: the "feature vector" that is used to computer node simularity/distance
        - `tag`: used to denote a group/category and map to a color
        - `source` and `bt`: which together denote the reference of the node element --
            `source` being an identification of the source of the data, and `bt` identifying (usually) time
            (or offset, or some addressing of the source stream).
        ... and any other fields, which will be ignored.
    :param nodeSize: The size of the displayed points (aka "nodes"), in pixels.
    :param height: Height of display rectangle, in pixels.
    :param width: Width of display rectangle, in pixels.
    :param untaggedColor: Color of an untagged node.
    :param maxIterations: Maximum iterations of the TSNE
    :param fps: Frames per second
    :param fillColors: List of colors to cycle through, one for every unique tag.
        How do `fillColors` and `tags` relate?
        Splatter iterates through the raw nodes to find unique tags.
        List of unique tags sorted in order tags were encountered.
        The mapping is then `..., unik_tag[i] -> fill_color[i], ...`.
    :param dim: Target dimension of the TSNE.
        If more than 2, only the first two dimensions are taken into account in the 2d visualization.
    :param epsilon: TSNE parameter. See https://distill.pub/2016/misread-tsne/
    :param perplexity: TSNE parameter. See https://distill.pub/2016/misread-tsne/
    :param spread: TSNE parameter. See https://distill.pub/2016/misread-tsne/
    :return:
    """
    kwargs = _splatter_raw_sig.extract_kwargs(*args, **kwargs)
    assert_jsonizable(kwargs)
    return _splatter(pts=kwargs.pop('pts'), options=kwargs)


def assert_pts_are_valid(pts):
    if not (isinstance(pts, list)  # pts are a list
            and len(pts) > 0  # with at least one element
            and isinstance(pts[0], dict)  # it's elements are dicts (at least the first)
            and 'fv' in pts[0]):  # that have an 'fv' field
        raise ValueError("pts must be a non-empty list of dicts that have at least an fv field")
    # All elements have an 'fv' and they are all of the same length
    first_fv_size = len(pts[0]['fv'])
    for pt in pts:
        if 'fv' not in pt:
            raise ValueError(f"An pt of pts didn't have an 'fv': {pt}")
        if not isinstance(pt['fv'], list):
            raise ValueError(f"All fvs of pts must be lists. This one was not: {pt}")
        if len(pt['fv']) != first_fv_size:
            raise ValueError(f"All fvs of pts must be of the same size ({first_fv_size}: {pt}")
    return True


# TODO: Forward JS errors to python and handle on python side (raising informative error for e.g.)
def _splatter(pts, options):
    assert_pts_are_valid(pts)
    if not options:
        options = {}
    return Javascript(f'splatter(element.get(0), {str(pts)}, {str(options)})')


# Just to note that we can do this with position only args too.

def call_func(func, **kwargs):
    args, kwargs = Sig(func).args_and_kwargs_from_kwargs(kwargs)
    return func(*args, **kwargs)


def call_func_ignoring_excess(func, **kwargs):
    """Call func, sourcing the arguments from kwargs and ignoring the excess arguments.
    Also works if func has some position only arguments.
    """
    s = Sig(func)
    args, kwargs = s.args_and_kwargs_from_kwargs(s.source_kwargs(**kwargs))
    return func(*args, **kwargs)

# TODO: Get defaults from splatter_defaults.json and inject in signature
# def splatter(pts,
#              nodeSize=1,
#              height=200,
#              width=200,
#              untaggedColor: str = '#444',
#              maxIterations=240,
#              fps=60,
#              fillColors=dflt_fillColors,
#              dim=2,
#              epsilon=50,
#              perplexity=30,
#              spread=10,
#              assert_same_sized_fvs: bool = True):
#     """
#     Splatter multidimensional pts (that is, see a TSNE iteration happen in front of your eyes,
#     squishing your multidimensional pts into two dimensions.
#
#     The `pts` input is a list of dicts, where every dict must have, at a minimum, an `'fv'` field whose value
#     is a list of fixed size for all pt of pts.
#
#     Optionally, you can include:
#     - 'tag': Will be use to categorize and color the point
#     - 'x': initial x-coordinate of the point
#     - 'y': initial y-coordinate of the point
#
#     ... and any other fields, which will be ignored.
#
#     :param pts: Your pts, in the form of a list of dicts, list of lists, or dict of lists
#     :param nodeSize: The size of the displayed points (aka "nodes")
#     :param height: Height of display rectangle
#     :param width: Width of display rectangle
#     :param untaggedColor: Color of an untagged node
#     :param maxIterations: Maximum iterations of the TSNE
#     :param fps: Frames per second
#     :param fillColors: List of colors to cycle through, one for every unique tag
#     :param dim: Target dimension of the TSNE
#     :param epsilon:
#     :param perplexity:
#     :param spread:
#     :param assert_same_sized_fvs:
#     :return:
#     """
#     fillColors = list(fillColors)  # because json can't handle tuples
#     options = dict(fillColors=fillColors, fps=fps, height=height, maxIterations=maxIterations, nodeSize=nodeSize,
#                    untaggedColor=untaggedColor, width=width, dim=dim, epsilon=epsilon, perplexity=perplexity,
#                    spread=spread)
#     pts = list(process_pts(pts))
#     return _splatter(pts, options, assert_same_sized_fvs)



from IPython.display import Javascript
import os
import soundfile

CHANNEL_TYPES = ['audio', 'data']


def mk_time_channel(filename, chart_type='peaks'):
    """
    Make a single_time_vis time channel dict from an audio file.

    :param filename: The filename to open
    """
    wf, sr = soundfile.read(filename, dtype='int16')
    return {
        'chartType': chart_type,
        'type': 'audio',
        'buffer': list(wf),
        'sr': sr
    }


def render_wav_file(filename,
                    chart_type='peaks',
                    enable_playback=True,
                    height=50,
                    params=None,
                    title='',
                    subtitle=''):
    """
    Renders a time visualization of a WAV file from its filename.

    :param filename: The filename to load
    :param chart_type: The chart type to render, either 'peaks' (default) or 'spectrogram
    :param enable_playback: Whether to enable playback on double click (default True)
    :param height: The height of the chart in pixels (default 50)
    :param params: Extra rendering parameters, currently unused
    :param title: The title to display, defaults to the filename
    :param subtitle: An optional subtitle to display under the title
    """
    channel = mk_time_channel(filename, chart_type)
    duration_s = len(channel['buffer']) / channel['sr']
    title = title or filename
    return single_time_vis(channel,
                           bt=0,
                           tt=int(duration_s * 1000000),
                           chart_type=chart_type,
                           enable_playback=enable_playback,
                           height=height,
                           params=params,
                           title=title,
                           subtitle=subtitle)


def single_time_vis(channel,
                    bt=0,
                    tt=10000000,
                    chart_type='bargraph',
                    enable_playback=True,
                    height=50,
                    params=None,
                    title='',
                    subtitle=''):
    """
    Render a visualization of time series or audio data.

    :param channel: A time channel dict (see below)
    :param bt: The logical start time of the display, in microseconds
    :param tt: The logical end time of the display, in microseconds
    :param chart_type: The type of visualization to draw, one of
        ["bargraph", "heatmap", "spectrogram", "peaks", or "winners"]
    :param enable_playback: Whether to embed audio controls in the component (audio channel only)
    :param height: The height of the channel in pixels
    :param params: An empty dict or a dict with "chunk_size" as an integer in milliseconds
    :param title: A title to display above the chart
    :param subtitle: A subtitle to display under the title

    Channel keys:

    Audio channel:
    Either "url" or "buffer"

    :param type: Must be "audio"
    :param url: A URL to access WAV-format audio over HTTP
    :param buffer: A list of 16-bit integers

    Data channel:

    :param type: Must be "data"
    :param data: A list of data points, with the keys "value" and either ("bt" and "tt") or "time"
    :param bargraph_max: The numeric value for the top of the chart, default 1
    :param bargraph_min: The numeric value for the top of the chart, default -2
    :param categories: For a winners channel, the list of categories to display
        (matching the "winners" values of the data points)
    """
    assert channel.get('type', None) in CHANNEL_TYPES, 'Channel type must be "audio" or "data"'
    if channel['type'] == 'audio':
        assert 'url' in channel or ('buffer' in channel and 'sr' in channel), 'Channel must have either "url" or "buffer" (list of integers) and "sr"'
        if not chart_type:
            chart_type = 'peaks'
    else:
        assert 'data' in channel
    props = dict({}, bt=bt, tt=tt, chart_type=chart_type, enable_playback=enable_playback, height=height,
                 params=params, title=title, subtitle=subtitle)
    return _single_time_vis(channel, props)


def time_vis(channels):
    """
    Render multiple channels.

    :param channels: The channels to render, a list of dicts.
    :return:
    """
    pass
    # options = dict({}, fillColors=fillColors, fps=fps, height=height, maxIterations=maxIterations, nodeSize=nodeSize,
    #                untaggedColor=untaggedColor, width=width, dim=dim, epsilon=epsilon, perplexity=perplexity,
    #                spread=spread)
    # return _splatter(pts, options, assert_same_sized_fvs)


def _single_time_vis(channel, props):
    js_source = f'renderTimeChannel(element.get(0), {channel}, {props})'.replace('True', 'true').replace('None', 'null')
    return Javascript(js_source)
